/* eval.c v1.0 - 2011-05-01
The code is based on similar codes by Herbert Schildt and Mark Morley (1991),
but with a number of corrections, additions and modifications.

Licence: Common Public License Version 1.0 (CPL)
  http://www.opensource.org/licenses/cpl1.0.php

Changelog:
  1.10 - 2011-05-07 :
    * fix sign, null line
    * result with 10 digits
    * add ClearSpaces()
    * clean code
  1.00 - 2011-05-01
    * initial version
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "eval.h"


void ERR (int n)
{
    ERROR = n;
    ERPOS = expression - ERANC - 1;
    strcpy (ERTOK, token);
}

TYPE deg (TYPE x)
{
    return (x * 180.0 / M_PI);
}

TYPE rad (TYPE x)
{
    return (x * M_PI / 180.0);
}

TYPE sign (TYPE x)
{
    if (x >= 0.0)
        return  1.0;
    else
        return -1.0;
}

TYPE cosdg (TYPE x)
{
    return cos (x * 180.0 / M_PI);
}

TYPE sindg (TYPE x)
{
    return sin (x * 180.0 / M_PI);
}

TYPE tandg (TYPE x)
{
    return tan (x * 180.0 / M_PI);
}

TYPE sqr (TYPE x)
{
    return x * x;
}

TYPE __jn (TYPE j, TYPE n)
{
    return jn ( (int) (j + 0.5), n);
}

TYPE __yn (TYPE y, TYPE n)
{
    return yn ( (int) (y + 0.5), n);
}

TYPE asinh (TYPE x)
{
    return log (x + sqrt (x * x + 1.0));
}

TYPE acosh (TYPE x)
{
    return log (x + sqrt (x * x - 1.0));
}

TYPE atanh (TYPE x)
{
    return log ((1.0 + x) / (1.0 - x)) / 2.0;
}

// sec, cosec, cot and their inverses

TYPE sec (TYPE x)
{
    return (1.0 / cos (x));
}

TYPE cosec (TYPE x)
{
    return (1.0 / sin (x));
}

TYPE cot (TYPE x)
{
    return (1.0 / tan (x));
}

TYPE asec (TYPE x)
{
    return acos (1.0 / x);
}

TYPE acosec (TYPE x)
{
    return asin (1.0 / x);
}

TYPE acot (TYPE x)
{
    return atan (1.0 / x);
}

// sech, cosech, coth and their inverses

TYPE sech (TYPE x)
{
    return (1.0 / cosh (x));
}

TYPE cosech (TYPE x)
{
    return (1.0 / sinh (x));
}

TYPE coth (TYPE x)
{
    return (1.0 / tanh (x));
}

TYPE asech (TYPE x)
{
    return acosh (1.0 / x);
}

TYPE acosech (TYPE x)
{
    return asinh (1.0 / x);
}

TYPE acoth (TYPE x)
{
    return atanh (1.0 / x);
}

TYPE log2 (TYPE x)
{
 return log (x) / log (2.0);
}

TYPE _round (TYPE x)
{
#ifdef round
    return round (x);
#else
    if (x < 0.0) return ceil  (x - 0.5);
    if (x > 0.0) return floor (x + 0.5);
    return 0.0;
#endif
}

TYPE _modf (TYPE x)
{
    TYPE retval;
    modf (x, &retval);
    return retval;
}

TYPE _modf2 (TYPE x)
{
    TYPE retval;
    return modf (x, &retval);
}

TYPE _trunc (TYPE x)
{
#ifdef trunc
    return trunc (x);
#else
    TYPE retval;
    modf (x, &retval);
    return retval;
#endif
}


/* This routine obtains a value from the program's environment. */
int GetSymbol (char* s, TYPE* v)
{
    char* e = getenv (s);

    if (!e)
        return 0;
    *v = atof (e);
    return 1;
}  /* GetSymbol */


/* Erases all user-defined variables from memory.
Note that constants can not be erased or modified in any way by the user. */
void ClearAllVars (void)
{
    int i;

    for (i = 0; i < MAXVARS; i++)
    {
        *(Vars[i].name) = '\0';
        Vars[i].value   = 0.0;
    }
}  /* ClearAllVars */


/* Erases the user-defined variable that is called NAME from memory.
Note that constants are not affected.
Returns 1 if the variable was found and erased, or 0 if it didn't exist. */
int ClearVar (char* name)
{
    int i;

    for (i = 0; i < MAXVARS; i++)
        if (*Vars[i].name && ! strcmp (name, Vars[i].name))
        {
            *(Vars[i].name) = '\0';
            Vars[i].value   = 0.0;
            return 1;
        }

    return 0;
}  /* ClearVar */


/* Looks up the specified variable (or constant) known as NAME and
returns its contents in VALUE. First the user-defined variables are searched,
then the constants are searched.
Returns 1 if the value was found, or 0 if it wasn't. */
int GetValue (char* name, TYPE* value)
{
    int i;

    /* First check for an environment variable reference... */
    if (*name == '_')
        return GetSymbol (name + 1, value);

    /* Now check the user-defined variables. */
    for (i = 0; i < MAXVARS; i++)
        if (*Vars[i].name && ! strcmp (name, Vars[i].name))
        {
            *value = Vars[i].value;
            return 1;
        }

    /* Now check the programmer-defined constants. */
    for (i = 0; *Consts[i].name; i++)
        if (*Consts[i].name && ! strcmp (name, Consts[i].name))
        {
            *value = Consts[i].value;
            return 1;
        }

    return 0;
}  /* GetValue */


/* First, it erases any user-defined variable that is called NAME.
Then it creates a new variable called NAME and gives it the value VALUE.
Returns 1 if the value was added, or 0 if there was no more room. */
int SetValue (char* name, TYPE* value)
{
    int i;

    ClearVar (name);
    for (i = 0; i < MAXVARS; i++)
        if (! *Vars[i].name)
        {
            strcpy (Vars[i].name, name);
            Vars[i].name[VARLEN] = '\0';
            Vars[i].value = *value;
            return 1;
        }

    return 0;
}  /* SetValue */


/* This function is used to grab the next token from the expression that is being evaluated. */
void Parse (void)
{
    char* t = token;
    type = 0;

    if (isdelim (*expression))
    {
        type = DEL;
        *t++ = *expression++;
    }
    else if (isnumer (*expression))
    {
        type = NUM;
        while (isnumer (*expression))
            *t++ = *expression++;
    }
    else if (isalpha (*expression))
    {
        type = VAR;
        while (isalpha (*expression))
            *t++ = *expression++;
        token[VARLEN] = 0;
    }
    else if (*expression)
    {
        *t++ = *expression++;
        *t = '\0';
        ERR (E_SYNTAX);
    }
    *t = '\0';
}  /* Parse */


/* This function handles any variable assignment operations.
It returns a value of 1 if it is a top-level assignment operation, otherwise it returns 0 */
int Level1 (TYPE* r)
{
    char t[VARLEN + 1];

    if (type == VAR)
        if (*expression == '=')
        {
            strcpy (t, token);
            Parse ();
            Parse ();
            if (!*token)
            {
                ClearVar (t);
                return 1;
            }
            Level2 (r);
            if (! SetValue (t, r))
                ERR (E_MAXVARS);
            return 1;
        }
    Level2 (r);

    return 0;
}  /* Level1 */


/* This function handles any addition and subtraction operations. */
void Level2 (TYPE* r)
{
    TYPE t = 0;
    char o;

    Level3 (r);
    while ( (o = *token) == '+' || o == '-')
    {
        Parse ();
        Level3 (&t);
        if (o == '+')
            *r = *r + t;
        else if (o == '-')
            *r = *r - t;
    }
}  /* Level2 */


/* This function handles any multiplication, division, or modulo. */
void Level3 (TYPE* r)
{
    TYPE t;
    char o;

    Level4 (r);
    while ( (o = *token) == '*' || o == '/' || o == '%')
    {
        Parse ();
        Level4 (&t);
        if (o == '*')
            *r = *r * t;
        else if (o == '/')
        {
            if (t == 0)
                ERR (E_DIVZERO);
            *r = *r / t;
        }
        else if (o == '%')
        {
            if (t == 0)
                ERR (E_DIVZERO);
            *r = fmod (*r, t);
        }
    }
}  /* Level3 */


/* This function handles any "to the power of" operations. */
void Level4 (TYPE* r)
{
    TYPE t;

    Level5 (r);
    if (*token == '^')
    {
        Parse ();
        Level5 (&t);
        *r = pow (*r, t);
    }
}  /* Level4 */


/* This function handles any unary + or - signs. */
void Level5 (TYPE* r)
{
    char o = 0;

    if (*token == '+' || *token == '-')
    {
        o = *token;
        Parse ();
    }
    Level6 (r);
    if (o == '-')
        *r = -*r;
}  /* Level5 */


/* This function handles any literal numbers, variables, or functions. */
void Level6 (TYPE* r)
{
    int  i, n;
    TYPE a[3];

    if (*token == '(')
    {
        Parse();
        if (*token == ')')
            ERR (E_NOARG);
        Level1 (r);
        if (*token != ')')
            ERR (E_UNBALAN);
        Parse ();
    }
    else
    {
        if (type == NUM)
        {
            *r = (TYPE) atof (token);
            Parse();
        }
        else if (type == VAR)
        {
            if (*expression == '(')
            {
                for (i = 0; *Funcs[i].name; i++)
                    if (! strcmp (token, Funcs[i].name))
                    {
                        Parse ();
                        n = 0;
                        do
                        {
                            Parse ();
                            if (*token == ')' || *token == ',')
                                ERR (E_NOARG);
                            a[n] = 0;
                            Level1 (&a[n]);
                            n++;
                        }
                        while (n < 4 && *token == ',');
                        Parse ();
                        if (n != Funcs[i].args)
                        {
                            strcpy (token, Funcs[i].name);
                            ERR (E_NUMARGS);
                        }
                             if (n == 1) *r = Funcs[i].func1 (a[0]);
                        else if (n == 2) *r = Funcs[i].func2 (a[0], a[1]);
                        return;
                    }
                if (! *Funcs[i].name)
                    ERR (E_BADFUNC);
            }
            else if (! GetValue (token, r))
                ERR (E_UNKNOWN);
            Parse();
        }
        else
            ERR (E_SYNTAX);
    }
}  /* Level6 */


/* This function is called to evaluate the expression E and
return the answer in RESULT. If the expression was a top-level assignment,
a value of 1 will be returned in A, otherwise it will contain 0.
Returns E_OK if the expression is valid, or an error code. */
int Evaluate (char* e, TYPE *result, int* a)
{
    ERANC = expression = e;
    *result = 0.0;
    Parse ();
    if (! *token)
        ERR (E_EMPTY);
    *a = Level1 (result);

    return E_OK;
}  /* Evaluate */


void PrintError (void)
{
    printf ("Error: %s - %s\n", ErrMsgs[ERROR - 1], ERTOK);
    puts   (ERANC);
    printf ("%*s^\n", ERPOS, "");
}  /* PrintError */


void PrintHelp (void)
{   /*
    puts (PRG_NAME" v"PRG_VERSION" - ("PRG_DATE")\n"
          "License: "PRG_LICENSE"\n"); */
    puts ("Commands:\n"
          "  cst : View contantes list\n"
          "  var : View variables list\n"
          "  cmd : View functions list\n"
          "  clr : Clear all variables\n"
          "  cls : Clear screen\n"
          "  help / ? : View this help\n\n"
          "Enter 'exit', 'quit' to close the program.\n"
          );
}  /* PrintHelp */


void PrintCmd (void)
{
    int i, n;
    char *name, spc[2] = " \0";
    puts ("Functions:");
    for (i = 0; *(name = Funcs[i].name); i++)
    {
        n = strlen (name);
             if (n <= 5) spc[0] = '\t';
        else if (n == 6) spc[0] = '\0';
        else if (n >= 7) spc[0] = '\b';

        if (Funcs[i].args == 1)
            printf ("  %s%s (A)    : %s\n", name, spc, Funcs[i].info);
        else if (Funcs[i].args == 2)
            printf ("  %s%s (A, B) : %s\n", name, spc, Funcs[i].info);
    }
}  /* PrintCmd */


char *ClearSpaces (char *szStr)
{
  int i, p = 0;
  char c;
  if ( szStr == NULL) return NULL;
  if (*szStr == '\0') return NULL;

  for (i = 0; (c = szStr[i]) != '\0'; i++)
  {
    if ((c != ' ') && (c != '\t') && (c != '\n') && (c != '\r'))
      szStr[p++] = c;
  }
  if (p != i) szStr[p] = '\0';

  return szStr;
}  /* ClearSpaces */


int main (int argc, char* argv[])
{
    TYPE  result;
    int   i, ec, a;
    char  line[1024];

    ClearAllVars ();

    if (argc > 1)  /* If we have command line arguments then we evaluate them and exit. */
    {
        /* Concatenate all arguments into one string. */
        strcpy (line, argv[1]);
        for (i = 2; i < argc; i++)
            strcat (line, argv[i]);

        /* Call the evaluator. */
        if ( (ec = Evaluate (line, &result, &a)) == E_OK)
        {
            if (!a)  /* If we didn't assign a variable, then print the result. */
                printf ("%.10g\n", result);
        }
        else if (ec != E_EMPTY)  /* Display error info. In this example, an E_EMPTY error is ignored. */
            PrintError ();
        return 0;
    }

    /* There were no command line arguments, so lets go interactive. */
    puts (PRG_NAME" v"PRG_VERSION" - ("PRG_DATE")\n"
          "License: "PRG_LICENSE"\n"
          "View help: 'help' or '?' - to exit: 'exit', 'quit'.\n");
    printf ("> ");

    /* Input one line at a time from the user.  Note that it uses stdin, so on
       DOS or UNIX you could pipe a list of expressions into it... */
    for (gets (line); !feof (stdin); gets (line))
    {
        ClearSpaces (line);
        strlwr (line);

        /* Did the user ask to exit? */
        if (!strcmp (line, "exit") || !strcmp (line, "quit"))
            return 0;

        /* Did the user ask to see the variables in memory? */
        else if (! strcmp (line, "var"))
        {
            for (i = 0; i < MAXVARS; i++)
                if (*Vars[i].name)
                    printf ("%s = %.10g\n", Vars[i].name, Vars[i].value);
        }

        /* Did the user ask to see the constants in memory? */
        else if (! strcmp (line, "cst"))
        {
            for (i = 0; *Consts[i].name; i++)
                printf ("%s\t%s= %.10g\n", Consts[i].name, (strlen (Consts[i].name) < 8) ? "\t" : "",  Consts[i].value);
        }

        else if (!strcmp (line, "help") || !strcmp (line, "?"))
            PrintHelp ();

        else if (!strcmp (line, "cmd"))
            PrintCmd ();

        /* Did the user ask to clear all variables? */
        else if (! strcmp (line, "clr"))
            ClearAllVars ();

        else if (! strcmp (line, "cls"))
            system ("CLS");

        else if (strlen (line))  /* If none of the above, then we attempt to evaluate the user's input. */
        {
            /* Call the evaluator. */
            if ( (ec = Evaluate (line, &result, &a)) == E_OK)
            {
                if (!a)  /* Only display the result if it was not an assignment. */
                    printf ("= %.10g\n", result);
            }
            else if (ec != E_EMPTY)  /* Display error information. E_EMPTY is ignored. */
                PrintError ();
        }
        printf ("> ");
    }

    return 0;
}  /* main */
