/* eval.h v1.0 - 2011-05-01
The code is based on similar codes by Herbert Schildt and Mark Morley (1991),
but with a number of corrections, additions and modifications.

License: Public Domain Dedication (CC0 1.0)
  http://creativecommons.org/publicdomain/zero/1.0/
*/

// http://pubs.opengroup.org/onlinepubs/007908799/xsh/math.h.html
// http://www.oschina.net/code/explore/mingw-runtime-3.18-1/mingwex/math/truncl.c
// http://www.linux-kheops.com/doc/man/manfr/man-html-0.9/man3/index.html

#ifndef _EVAL_H_
#define _EVAL_H_

#include <math.h>
#include <float.h>

#include "cbrt.h"

#define PRG_NAME        "Eval"
#define PRG_VERSION     "1.10"
#define PRG_DATE        "2011-05-07"
#define PRG_LICENSE     "Public Domain Dedication (CC0 1.0)"

#define TYPE            double      /* Type of numbers to work with */

#define VARLEN          15          /* Max length of variable names */
#define MAXVARS         50          /* Max user-defined variables */
#define TOKLEN          30          /* Max token length */

#define VAR             1
#define DEL             2
#define NUM             3

#ifndef M_PI
#define M_PI		    3.14159265358979323846
#define M_E		        2.7182818284590452354
#define M_SQRT2		    1.41421356237309504880
#endif

#define MAX_DWORD_D     4294967295.0
#define MAX_WORD_D      65535.0
#define MAX_INT_D       2147483647.0
#define MIN_INT_D       -2147483648.0

#ifndef rint
#define rint            _round
#endif

#define neg             _chgsign

typedef struct
{
   char name[VARLEN + 1];    /* Variable name */
   TYPE value;               /* Variable value */
} VARIABLE;

typedef struct
{
   char *name;               /* Function name */
   int  args;                /* Number of arguments to expect */
   TYPE (*func1)(TYPE);      /* Pointer to function */
   TYPE (*func2)(TYPE,TYPE);
   const char *info;
} FUNCTION;

/* The following macros are ASCII dependant, no EBCDIC here! */
#define isnumer(c)  ((c >= '0' && c <= '9') ||  c == '.')
#define isalpha(c)  ((c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c == '_')
#define isdelim(c)   (c == '+' || c == '-'  ||  c == '*' || c == '/'  || c == '%' || \
                      c == '^' || c == '('  ||  c == ')' || c == ','  || c == '=')

/* Codes returned from the evaluator */
#define E_OK           0     /* Successful evaluation */
#define E_SYNTAX       1     /* Syntax error */
#define E_UNBALAN      2     /* Unbalanced parenthesis */
#define E_DIVZERO      3     /* Attempted division by zero */
#define E_UNKNOWN      4     /* Reference to unknown variable */
#define E_MAXVARS      5     /* Maximum variables exceeded */
#define E_BADFUNC      6     /* Unrecognised function */
#define E_NUMARGS      7     /* Wrong number of arguments to funtion */
#define E_NOARG        8     /* Missing an argument to a funtion */
#define E_EMPTY        9     /* Empty expression */

int   ERROR;                 /* The error number */
char  ERTOK[TOKLEN + 1];     /* The token that generated the error */
int   ERPOS;                 /* The offset from the start of the expression */
char* ERANC;                 /* Used to calculate ERPOS */

/* What follows is a main() routine that evaluates the command line
arguments one at a time, and displays the results of each expression.
Without arguments, it becomes an interactive calculator. */
const char *ErrMsgs[] =
{
    "Syntax error",
    "Unbalanced parenthesis",
    "Division by zero",
    "Unknown variable",
    "Maximum variables exceeded",
    "Unrecognised funtion",
    "Wrong number of arguments to funtion",
    "Missing an argument",
    "Empty expression"
};

/* Add any "constants" here...  These are "read-only" values that are provided
as a convienence to the user. Their values can not be permanently changed.
The first field is the variable name, the second is its value. */
VARIABLE Consts[] =
{
   /* name,             value */
   { "Pi"             , M_PI           },
   { "e"              , M_E            },
   { "ans"            , 0.0            },
   { "BoltzmannConst" ,	1.3806505e-23  },
   { "ElectronCharge" , 1.60217653e-19 },
   { "NA"             , 6.02214199e23  },
   { "e0"             , 8.854187817e-12},
   { "mu0"            , M_PI*4e-7      },
   { "VelocityOfLight",	299792458      },
   { "PlankConst"     , 6.62606896e-34 },
   { "ElectronMass"   ,	9.1093826e-31  },
   { "ProtonMass"     , 1.67262171e-27 },
   { "Infinity"       , DBL_MAX        },
   { "_Infinity"      , DBL_MIN        },
   { "Laplace"        , 0.6627434193   },
   { "NumGold"        , 1.61803398874989484820},
   { "Sq2"            , M_SQRT2        },  /* = sqrt (2.0) */
   { "mDW"            , MAX_DWORD_D    },  /* =  2^32 - 1  */
   { "mW"             , MAX_WORD_D     },  /* =  2^16 - 1  */
   { "mI"             , MAX_INT_D      },  /* =  2^31 - 1  */
   { "_mI"            , MIN_INT_D      },  /* = -2^31      */
   { ""               , 0.0            }
};


VARIABLE  Vars[MAXVARS];       /* Array for user-defined variables */
char*     expression;          /* Pointer to the user's expression */
char      token[TOKLEN + 1];   /* Holds the current token */
int       type;                /* Type of the current token */

TYPE deg     (TYPE x);
TYPE rad     (TYPE x);

TYPE sign    (TYPE x);
TYPE cosdg   (TYPE x);
TYPE sindg   (TYPE x);
TYPE tandg   (TYPE x);
TYPE sqr     (TYPE x);
TYPE __jn    (TYPE j, TYPE n);
TYPE __yn    (TYPE y, TYPE n);
TYPE log2    (TYPE x);
TYPE _round  (TYPE x);
TYPE _modf   (TYPE x);
TYPE _modf2  (TYPE x);
TYPE _trunc  (TYPE x);

TYPE asinh   (TYPE x);
TYPE acosh   (TYPE x);
TYPE atanh   (TYPE x);

TYPE sec     (TYPE x);
TYPE cosec   (TYPE x);
TYPE cot     (TYPE x);
TYPE asec    (TYPE x);
TYPE acosec  (TYPE x);
TYPE acot    (TYPE x);
TYPE sech    (TYPE x);
TYPE cosech  (TYPE x);
TYPE coth    (TYPE x);
TYPE asech   (TYPE x);
TYPE acosech (TYPE x);
TYPE acoth   (TYPE x);

/* Add any math functions that you wish to recognise here...
The first field is the name of the function as it would appear in an expression.
The second field tells how many arguments to expect.
The third is a pointer to the actual function to use. */
FUNCTION Funcs[] =
{
    /* name, funtion to call */
    { "sin"  ,   1,    sin  , NULL , "Sine"                         },
    { "cos"  ,   1,    cos  , NULL , "Cosine"                       },
    { "tan"  ,   1,    tan  , NULL , "Tangent"                      },
    { "asin" ,   1,    asin , NULL , "Arc sine"                     },
    { "acos" ,   1,    acos , NULL , "Arc cosine"                   },
    { "atan" ,   1,    atan , NULL , "Arc tangent"                  },

    { "exp"  ,   1,    exp  , NULL , "Exponential function base e"  },
    { "ln"   ,   1,    log  , NULL , "Logarithm base e"             },
    { "log"  ,   1,    log10, NULL , "Logarithm base 10"            },
    { "sqrt" ,   1,    sqrt , NULL , "Square root"                  },
    { "floor",   1,    floor, NULL , "Computes nearest integer not greater than A"  },
    { "ceil" ,   1,    ceil , NULL , "Computes nearest integer not less than A"     },
    { "abs"  ,   1,    fabs , NULL , "Absolute value of an floating point value"    },

    { "deg"  ,   1,    deg  , NULL , "Convert angle to Degre"       },
    { "rad"  ,   1,    rad  , NULL , "Convert angle to Radian"      },

    { "pow"  ,   2,    NULL , pow  , "Power or A^B"                 },
    { "atan2",   2,    NULL , atan2, "Arc tangent, using signs to determine quadrants"      },
    { "fmod" ,   2,    NULL , fmod , "Computes the remainder of the division operator A/B"  },
    { "hypot",   2,    NULL , hypot, "Euclidean distance function = sqrt (A+B)"           },

    { "sinh",    1,    sinh , NULL , "Hyperbolic sine"              },
    { "cosh",    1,    cosh , NULL , "Hyperbolic cosine"            },
    { "tanh",    1,    tanh , NULL , "Hyperbolic tangent"           },
    { "asinh",   1,    asinh, NULL , "Hyperbolic arc sine"          },
    { "acosh",   1,    acosh, NULL , "Hyperbolic arc cosine"        },
    { "atanh",   1,    atanh, NULL , "Hyperbolic arc tangent"       },

    { "sign"   , 1,    sign , NULL , "Signum"                       },
    { "cosdg"  , 1,    cosdg, NULL , "Cosine in degre"              },
    { "sindg"  , 1,    sindg, NULL , "Sinus in degre"               },
    { "tandg"  , 1,    tandg, NULL , "Tangent in degre"             },
    { "sqr"    , 1,    sqr  , NULL , "Square"                       },

    { "j0"     , 1,    j0     , NULL , "Bessel functions of A of the first kind of order 0" },
    { "j1"     , 1,    j1     , NULL , "Bessel functions of A of the first kind of order 1" },
    { "jn"     , 2,    NULL   , __jn , "Bessel functions of A of the first kind of order n" },
    { "y0"     , 1,    y0     , NULL , "Bessel functions of A of the second kind of order 0"},
    { "y1"     , 1,    y1     , NULL , "Bessel functions of A of the second kind of order 1"},
    { "yn"     , 2,    NULL   , __yn , "Bessel functions of A of the second kind of order n"},
    { "neg"    , 1,    neg    , NULL , "Reverses the sign"                                  },
    { "logb"   , 1,    _logb  , NULL , "Radix-independent exponent"                         },
    { "log2"   , 1,    log2   , NULL , "Logarithm base 2"                                   },
    { "round"  , 1,    _round , NULL , "round-to-nearest integral value"                    },
    { "rint"   , 1,    rint   , NULL , "round-to-nearest integral value"                    },
    { "trunc"  , 1,    _trunc , NULL , "Returns the integer part"                           },
    { "int"    , 1,    _modf  , NULL , "Returns the integer part"                           },
    { "frac"   , 1,    _modf2 , NULL , "Returns the fractional part"                        },
    { "cbrt"   , 1,    _cbrt  , NULL , "Cube root"                                          },

    { "sec"    , 1,    sec    , NULL , "Secans"                                             },
    { "cosec"  , 1,    cosec  , NULL , "Co-Secans"                                          },
    { "cot"    , 1,    cot    , NULL , "Co-Tangens = 1/tan"                                 },
    { "asec"   , 1,    asec   , NULL , "Arcus secans = inverse of sec"                      },
    { "acosec" , 1,    acosec , NULL , "Arcus co-secans = inverse of cosec"                 },
    { "acot"   , 1,    acot   , NULL , "Area-co-tangens hyperbolicus = inverse of coth"     },
    { "sech"   , 1,    sech   , NULL , "Secans hyperbolicus"                                },
    { "cosech" , 1,    cosech , NULL , "Co-Secans hyperbolicus"                             },
    { "coth"   , 1,    coth   , NULL , "Co-Tangens hyperbolicus"                            },
    { "asech"  , 1,    asech  , NULL , "Area-secans hyperbolicus = invers of sech"          },
    { "acosech", 1,    acosech, NULL , "Area-co-secans hyperbolicus = inverse of cosech"    },
    { "acoth"  , 1,    acoth  , NULL , "Area-co-tangens hyperbolicus = inverse of coth"     },
    { ""       , 0,    NULL   , NULL , ""                                                   }
};

int  Level1 (TYPE* r);
void Level2 (TYPE* r);
void Level3 (TYPE* r);
void Level4 (TYPE* r);
void Level5 (TYPE* r);
void Level6 (TYPE* r);

#endif  /* _EVAL_H_ */
